local timeout = tonumber(c8y.config:value("timeout", 60))

--TODO clean up code copying
function rdbGet(parameter)
  local file = io.popen("rdb get " .. parameter)
  local result
  if file then
    result = file:read("*l")
    file:close()
  end
  return result
end

function rdbSet(parameter, value)
  local file = io.popen("rdb set " .. parameter .. " " .. value)
  if file then
    file:close()
  end
end

function rdbSetFlags(parameter, flags)
  local file = io.popen("rdb setflags " .. parameter .. " " .. flags)
  if file then
    file:close()
  end
end

function findLinkProfile()
  local result
  for index=1, 6 do
    result = rdbGet("link.profile." .. index .. ".defaultroute")
    if result == "1" then
      return index
    end
  end
  return nil
end

function setPdp(onOff)
  local linkProfile = findLinkProfile()
  if linkProfile then
    print("SETTING PDP TO " .. onOff)
    rdbSet("link.profile." .. linkProfile .. ".enable", onOff)
    return true
  end
end

function setConnectionMode(mode)
  rdbSet("service.cumulocity.plugin.lua__smsSwitch.mode", mode)
end

function getConnectionMode()
  return rdbGet("service.cumulocity.plugin.lua__smsSwitch.mode")
end
-------------------------------------------------------------------
class("ModeSetter")

function ModeSetter:operationExecuting(op)
  local mode = op:getResponseRecord(1)[3]
  if mode:lower() == "sms" then
    rdbSet("service.cumulocity.sms.enable",1)
    rdbSetFlags("service.cumulocity.sms.enable", "p")
    local startTime = os.time()
    local opId = op.id
    c8y.timer.start(5, function(timer)
      local confirm = getConnectionMode()
      if confirm == "sms" then
        print("SWITCHING TO SMS MODE CONFIRMED")
        op:stopTimeout()
        timer:stop()
      elseif os.difftime(os.time(),startTime) > timeout then
        print("SWITCHING TO SMS MODE FAILED")
        rdbSet("service.cumulocity.sms.enable",0)
        c8y.operations:setOperationStatus(opId, c8y.operations.Failed, "Switching to SMS mode failed.")
        timer:stop()
      end
    end)
  else
    op:setStatus(op.Failed, "Switching to communcation mode " .. (mode or "nil") .. " not supported.")
  end
end
-------------------------------------------------------------------

local	oldSmsEnable = "no value"

function watchSmsAgent()
  print("START WATCHING SMS AGENT")
  c8y.timer.start(5, function(timer)
    local smsEnable = rdbGet("service.cumulocity.sms.enable")
    local confirm = rdbGet("service.cumulocity.modeswitching.confirm")
    if smsEnable ~= oldSmsEnable then
      if smsEnable ~= "1" then
        print("SMS AGENT TERMINATED")
        setPdp(1)
        setConnectionMode("ip")
        c8y.timer.singleShot(60, function()
            c8y.operations:notificationsOn()
          end)
        oldSmsEnable = smsEnable
      elseif confirm == "1" then
        print("SMS AGENT START CONFIRMED")
        c8y.operations:notificationsOff()
        setPdp(0)
        rdbSet("service.cumulocity.modeswitching.confirm", 0)
        setConnectionMode("sms")
        oldSmsEnable = smsEnable
      end
    end
  end)
end

function init()
  local modeSetter = ModeSetter()
  c8y.operations:addOperation("c8y_CommunicationMode", "$.mode", modeSetter)

  watchSmsAgent()
end
